
#define _WIN32_WINNT 0x400
#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>

#include <string>
#include <vector>
#include <list>
#include <time.h>
#include <algorithm>
#include <fstream>
#include <io.h>
#include <map>

using std::list;
using std::vector;
using std::string;

#include "D3DApp.h"
#include "common_globals.h"
#include "shader.h"
#include "deferred.h"

#include "EffectLayout.h"

#include "effect.h"
#include "EffectShade.h"

void EffectShade::Init() {
  if (!m_shader) {
    m_shader = new Shader();
  }
  const EffectParam *ep;
  ep = GetPR()->get("shade");
  m_shader->CreateFromFile(g_D3DApp->m_pd3dDevice, ep->getString("file"));
}
void EffectShade::ReloadShaders() {
  Init();
}
bool EffectShade::CheckShadersModifiedOnDisk() {
  if (m_shader && m_shader->CheckModifiedOnDisk()) {
    return true;
  }
  return false;
}

void EffectShade::Advance() {
}

// renders all stuff belonging to the effect, can be called multiple times per frame
int EffectShade::Render() {
  // piirtoa

  LPDIRECT3DDEVICE9 pd3dDevice = g_D3DApp->m_pd3dDevice;

  const EffectParam *ep;
  ep = GetPR()->get("shade");


  pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW); 
  pd3dDevice->SetRenderState(D3DRS_LIGHTING, false);
  pd3dDevice->LightEnable(0, false);
  pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, false);

  pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
  pd3dDevice->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
  pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
  pd3dDevice->SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
  pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG0, D3DTA_CURRENT);
 
  pd3dDevice->SetRenderState(D3DRS_TEXTUREFACTOR, 0xFFFFFFFF);

  pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
  pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

  pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_BLENDDIFFUSEALPHA);
  pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
  pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

  pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
  pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

  pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

  pd3dDevice->SetRenderState(D3DRS_ZENABLE, true);
  pd3dDevice->SetRenderState(D3DRS_ZWRITEENABLE, true);
  pd3dDevice->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);

 // Texture *rt16b_buffer0 = g_D3DApp->createRenderTarget(g_D3DApp->m_windowWidth, g_D3DApp->m_windowHeight, "rt16b_buffer0", 1, D3DFMT_A16B16G16R16F);
  Texture *rt16b_depth = GetDeferred()->GetRT16bDepth();
  Texture *rt16b_diffuse = GetDeferred()->GetRT16bDiffuse();
  Texture *rt16b_normal = GetDeferred()->GetRT16bNormal();
  Texture *rt16b_ambSpec = GetDeferred()->GetRT16bAmbSpec();



  HRESULT hr = pd3dDevice->BeginScene();

  if (FAILED(hr)) {
    return hr;
  }

  pd3dDevice->SetRenderState(D3DRS_ZWRITEENABLE, false);
  pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_POINT );
  pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_POINT );

  pd3dDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
  pd3dDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);


  D3DXMATRIXA16 matProj;
  D3DXMATRIXA16 matWorld;
  D3DXMATRIXA16 matWVP;

  D3DXMatrixIdentity(&matProj);
  pd3dDevice->SetTransform(D3DTS_PROJECTION, &matProj);
  pd3dDevice->SetTransform(D3DTS_VIEW, &matProj);
  pd3dDevice->SetTransform(D3DTS_WORLD, &matProj);
  D3DXMatrixIdentity(&matWorld);

  matWVP = matWorld*matProj*matProj;

  m_shader->GetEffect()->SetMatrix( "g_mWorldViewProjection", &matWVP );

  m_shader->GetEffect()->SetFloat( "g_windowWidth", g_D3DApp->m_windowWidth);
  m_shader->GetEffect()->SetFloat( "g_windowHeight", g_D3DApp->m_windowHeight);

  int shadeNum = GetPR()->getN("shade");

  for (int ei=0; ei<shadeNum; ei++) {
    ep = GetPR()->get("shade", ei);

   //  g_D3DApp->setTexture(rt16b_buffer0);
    int bUseDiffuse = ep->getInt("diffuse_on");
    int bUseDepth = ep->getInt("depth_on");
    int bUseNormal = ep->getInt("normal_on");
    int bUseAmbSpec = ep->getInt("ambspec_on");

    int bUseBlend = ep->getInt("blend_on");

    m_shader->GetEffect()->SetFloat("g_time", m_effectTime);

    string blendMode = "";
    if (ep->getN("blend_mode"))
      blendMode = ep->getString("blend_mode");


    if (bUseDiffuse) {
      m_shader->GetEffect()->SetTexture("g_tDiffuse", rt16b_diffuse->lpTexture);
    }
    if (bUseDepth) {
      m_shader->GetEffect()->SetTexture("g_tDepth", rt16b_depth->lpTexture);
    }
    if (bUseNormal) {
      m_shader->GetEffect()->SetTexture("g_tNormal", rt16b_normal->lpTexture);
    }
    if (bUseAmbSpec) {
      m_shader->GetEffect()->SetTexture("g_tAmbSpec", rt16b_ambSpec->lpTexture);
    }

    int numParams;

    numParams = ep->getN("texture");
    for (int param=0; param<numParams; param++) {
      const EffectParam *epp = ep->get("texture", param);
      std::string name = epp->getString("shader_name");
      std::string file = epp->getString("file");
      Texture *tex = g_D3DApp->addTexture(file.c_str());
      if (tex) {
        m_shader->GetEffect()->SetTexture(name.c_str(), tex->lpTexture);
      }
    }

    numParams = ep->getN("param");
    for (int param=0; param<numParams; param++) {
      const EffectParam *epp = ep->get("param", param);
      std::string name = epp->getString("shader_name");
      D3DXVECTOR4 value = epp->getVec4("value");
      m_shader->GetEffect()->SetVector(name.c_str(), &value);
    }

    std::string tech = "Render";
    if (ep->getN("tech")) {
      tech = ep->getString("tech");
    }

    if (bUseBlend) {
      pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
    } else {
      pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, false);
    }

    if (strcmp(blendMode.c_str(), "add")==0) {
      pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
      pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    } else {
      pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
      pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    }


    m_shader->Enable(tech.c_str());

    m_shader->Flush();
    g_D3DApp->drawSquare(1.0f);

    m_shader->Disable();


  }


  // End the scene.
  pd3dDevice->EndScene();

  return 1;
}
